// Copyright 2023 The envd Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package compileui

import (
	"io"
	"time"
)

type trace struct {
	w           io.Writer
	startTime   *time.Time
	modeConsole bool
}

func newTrace(w io.Writer, modeConsole bool) *trace {
	return &trace{
		w:           w,
		modeConsole: modeConsole,
	}
}

func (t *trace) init() {
	current := time.Now()
	t.startTime = &current
}
