// Copyright 2023 The envd Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package syncthing

import (
	"net/url"
	"strings"
)

func ParsePortFromAddress(addr string) string {
	if strings.Contains(addr, "://") {
		rawUrl, err := url.Parse(addr)
		if err != nil {
			return ""
		}
		return rawUrl.Port()
	}

	splitLst := strings.Split(addr, ":")
	if len(splitLst) == 2 {
		return splitLst[1]
	}

	return ""
}
